@@ -1,6 +1,10 @@
+import collections
 import json
 import logging
+import re
+import string
 import subprocess
+from functools import partial
 from io import BytesIO, StringIO
 from pathlib import Path
 from typing import Type
@@ -129,7 +133,19 @@ class HasFormException(Exception):
     pass
 
 
-class NoFontException(Exception):
+class NoTextException(Exception):
+    pass
+
+
+class AlreadyProcessed(Exception):
+    pass
+
+
+class TooManyFontsException(Exception):
+    pass
+
+
+class TooManyPagesException(Exception):
     pass
 
 
@@ -143,11 +159,10 @@ def has_form(qpdf_page, operands):
     return False
 
 
-def parse_text(qpdf_page: pikepdf.Page, font_map, synthesizer_class: Type[PdfSynthesizer], gt):
+def parse_text(qpdf_page: pikepdf.Page, font_map, synthesizer: PdfSynthesizer):
     content_stream = iter(pikepdf.parse_content_stream(qpdf_page))
     new_content_stream = []
     last_used_font = None
-    synthesizer = synthesizer_class(gt, font_map)
 
     for operands, operator in content_stream:
         if operator == pikepdf.Operator('Do'):
@@ -171,37 +186,72 @@ def parse_text(qpdf_page: pikepdf.Page, font_map, synthesizer_class: Type[PdfSyn
         else:
             new_content_stream.append((operands, operator))
 
-    return new_content_stream, synthesizer.create_new_ground_truth()
+    return new_content_stream
 
 
-def synthesize_pdf(pdf_file, json_file, dst_dir, synthesizer_class):
+def synthesize_pdf(
+    pdf_file,
+    json_file,
+    dst_dir,
+    max_fonts,
+    max_pages,
+    num_outputs_per_document,
+    synthesizer_class,
+):
     ground_truth = json.loads(json_file.read_text())
     pdf_io = BytesIO(pdf_file.read_bytes())
     output_string = StringIO()
     rsrcmgr = PDFResourceManager(caching=True)
     device = TextConverter(rsrcmgr, output_string, codec='utf-8', laparams=LAParams())
     interpreter = PDFPageInterpreter(rsrcmgr, device)
+    interpreter_fonts = {}
+
+    def _out_path(_i, suffix):
+        return dst_dir / f'{json_file.stem}-{_i}{suffix}'
+
+    k_to_process = []
+    for i in range(num_outputs_per_document):
+        if not (_out_path(i, '.pdf').exists() and _out_path(i, '.json').exists()):
+            k_to_process.append(i)
+
+    if not k_to_process:
+        raise AlreadyProcessed(f'Already processed {pdf_file} {json_file}')
 
     with pikepdf.Pdf.open(pdf_file) as pdf:
-        for page, miner in zip(pdf.pages, PDFPage.get_pages(pdf_io)):
+        if max_pages and len(pdf.pages) > max_pages:
+            raise TooManyPagesException(f'Too many pages {len(pdf.pages)} > {max_pages} in PDF, skipping!')
+
+        for page_number, (page, miner) in enumerate(zip(pdf.pages, PDFPage.get_pages(pdf_io))):
             interpreter.process_page(miner)
+            interpreter_fonts.update(interpreter.fontmap)
 
-            font_map = {}
-            for font_name, font_val in page.resources.get('/Font', {}).items():
-                if font_obj := interpreter.fontmap.get(font_name.lstrip('/')):
-                    font_map[font_name] = Font(font_name, font_obj)
+    if max_fonts and len(interpreter_fonts) > max_fonts:
+        raise TooManyFontsException(f'Too many fonts {len(interpreter_fonts)} > {max_fonts} in PDF, skipping!')
 
-            if not font_map:
-                raise NoFontException
+    if not re.sub(f'[{re.escape(string.whitespace)}]', '', output_string.getvalue()):
+        raise NoTextException('PDF does not have any text! Skipping')
 
-            new_content_stream, new_ground_truth = parse_text(page, font_map, synthesizer_class, ground_truth)
-            page.Contents = pdf.make_stream(pikepdf.unparse_content_stream(new_content_stream))
+    font_map = {f'/{k}': Font(f'/{k}', v) for k, v in interpreter_fonts.items()}
+    synthesizer = synthesizer_class(ground_truth, font_map)
 
-        out_dst = dst_dir / pdf_file.name
-        pdf.save(out_dst)
+    with pikepdf.Pdf.open(pdf_file) as pdf:
+        new_contents = collections.defaultdict(list)
+        new_ground_truths = []
 
-    out_json_dst = dst_dir / json_file.name
-    out_json_dst.write_text(json.dumps(new_ground_truth, indent=2))
+        for i in k_to_process:
+            for page_number, page in enumerate(pdf.pages):
+                new_content_stream = parse_text(page, font_map, synthesizer)
+                new_contents[i].append(pdf.make_stream(pikepdf.unparse_content_stream(new_content_stream)))
+
+            new_ground_truths.append(synthesizer.create_new_ground_truth())
+            synthesizer.reset()
+
+        for i in k_to_process:
+            for page_number, page in enumerate(pdf.pages):
+                page.Contents = new_contents[i][page_number]
+
+            pdf.save(_out_path(i, '.pdf'))
+            _out_path(i, '.json').write_text(json.dumps(new_ground_truths[i], indent=2))
 
 
 def parse_pdf(
@@ -209,30 +259,43 @@ def parse_pdf(
     pdf_file: Path,
     json_file: Path,
     synthesizer_class: Type[PdfSynthesizer],
+    max_fonts: int,
+    max_pages: int,
+    num_outputs_per_document: int,
     dst_dir: Path,
     tmp_dir: Path
 ):
     logger.info(f'{name}: {pdf_file} {json_file}')
     status = f'Error when synthesizing {name}'
+    synthesize_fn = partial(
+        synthesize_pdf,
+        json_file=json_file,
+        dst_dir=dst_dir,
+        max_fonts=max_fonts,
+        max_pages=max_pages,
+        num_outputs_per_document=num_outputs_per_document,
+        synthesizer_class=synthesizer_class,
+    )
 
     try:
-        synthesize_pdf(pdf_file, json_file, dst_dir, synthesizer_class)
+        synthesize_fn(pdf_file)
         status = f'Successfully synthesized {name}'
     except HasFormException:
         logger.info('Has form! Trying to flatten PDF')
-        flattened_pdf_file = flatten(pdf_file, tmp_dir)
-        synthesize_pdf(flattened_pdf_file, json_file, dst_dir, synthesizer_class)
-        flattened_pdf_file.unlink()
-        status = f'Successfully synthesized {name}'
-    except NoFontException:
-        logger.warning('PDF does not have any fonts! Skipping')
-    except FileNotFoundError as e:
+        if flattened_pdf_file := flatten(pdf_file, tmp_dir):
+            synthesize_fn(flattened_pdf_file)
+            flattened_pdf_file.unlink()
+            status = f'Successfully synthesized {name}'
+    except (AlreadyProcessed, FileNotFoundError, NoTextException, TooManyFontsException, TooManyPagesException) as e:
         logger.error(e)
 
     return status
 
 
 def flatten(pdf_file, tmp_dir):
-    dst_path = tmp_dir / f'{pdf_file.stem}.flattened.pdf'
-    subprocess.run(f'gs -q -sDEVICE=pdfwrite -o {dst_path} {pdf_file}', shell=True)
-    return dst_path
+    try:
+        dst_path = tmp_dir / f'{pdf_file.stem}.flattened.pdf'
+        subprocess.run(f'gs -q -sDEVICE=pdfwrite -o {dst_path} {pdf_file}', shell=True, timeout=15)
+        return dst_path
+    except (subprocess.CalledProcessError, subprocess.TimeoutExpired) as e:
+        logger.error(e)
